using System;
using System.ComponentModel;
using System.IO;
using System.Runtime.InteropServices;

namespace Lucidiot.CertBoat {

    internal static class NativeMethods {
        private const uint PKCS_7_ASN_ENCODING = 0x00010000;
        private const uint X509_ASN_ENCODING = 0x00000001;

        private enum CertStoreAddDisposition {
            New = 1,
            UseExisting = 2,
            ReplaceExisting = 3,
            Always = 4,
            ReplaceExistingInheritProperties = 5,
            AddNewer = 6,
            AddNewerInheritProperties = 7
        }

        [DllImport("crypt32.dll", SetLastError = true)]
        private static extern IntPtr CertOpenSystemStore(
            IntPtr hprov,
            [MarshalAs(UnmanagedType.LPStr)]
            string szSubsystemProtocol
        );

        [DllImport("crypt32.dll", SetLastError = true)]
        private static extern bool CertAddEncodedCTLToStore(
            IntPtr hCertStore,
            uint dwMsgAndCertEncodingType,
            [MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 3)]
            byte[] pbCtlEncoded,
            uint cbCtlEncoded,
            uint dwAddDisposition,
            IntPtr ppCtlContext
        );

        [DllImport("crypt32.dll", SetLastError = true)]
        private static extern bool CertCloseStore(
            IntPtr hCertStore,
            uint dwFlags
        );

        internal static void AddCTLToStore(byte[] ctlData) {
            IntPtr store = CertOpenSystemStore(IntPtr.Zero, Properties.Settings.Default.SystemStoreName);
            if (store == IntPtr.Zero)
                throw new Win32Exception(Marshal.GetLastWin32Error());

            if (!CertAddEncodedCTLToStore(
                store,
                X509_ASN_ENCODING | PKCS_7_ASN_ENCODING,
                ctlData,
                (uint)ctlData.Length,
                (uint)CertStoreAddDisposition.ReplaceExisting,
                IntPtr.Zero
            )) {
                /*
                 * If CertAddEncodedCTLToStore fails, we try to close the system store.
                 * If closing the system store fails too, we throw a Win32Exception with a Win32Exception inside!
                 */
                Win32Exception exc = new Win32Exception(Marshal.GetLastWin32Error());
                if (CertCloseStore(store, 0))
                    throw exc;
                else
                    throw new Win32Exception(Marshal.GetExceptionForHR(Marshal.GetLastWin32Error()).Message, exc);
            }

            if (!CertCloseStore(store, 0))
                throw new Win32Exception(Marshal.GetLastWin32Error());
        }

        internal static void AddCTLToStore(string path) {
            AddCTLToStore(File.ReadAllBytes(path));
        }
    }
}
